/* $Id: mapwms.c,v 1.99.2.3 2004/08/26 10:27:09 valik Exp $ */

#include "map.h"
#include "maperror.h"
#include "mapgml.h"
#include "gdfonts.h"

#include "maptemplate.h"

#include "mapogcsld.h"

#include <stdarg.h>
#include <time.h>
#include <string.h>

/*
** msWMSGetEPSGProj() moved to mapproject.c and renamed msGetEPSGProj(). This function turns
** out to be generally useful outside of WMS context.
*/

/* ==================================================================
 * WMS Server stuff.
 * ================================================================== */
#ifdef USE_WMS_SVR

/*
** msWMSException()
**
** Report current MapServer error in requested format.
*/
static char *wms_exception_format=NULL;

int msWMSException(mapObj *map, const char *wmtversion, const char *exception_code)
{
  // Establish default exception format depending on VERSION
  if (wms_exception_format == NULL)
  {
      if (wmtversion && strcasecmp(wmtversion, "1.0.0") <= 0)
        wms_exception_format = "INIMAGE";  // WMS 1.0.0
      else if (wmtversion && strcasecmp(wmtversion, "1.0.7") <= 0)
        wms_exception_format = "SE_XML";   // WMS 1.0.1 to 1.0.7
      else
        wms_exception_format = "application/vnd.ogc.se_xml"; // WMS 1.0.8, 1.1.0 and later
  }

  if (strcasecmp(wms_exception_format, "INIMAGE") == 0 ||
      strcasecmp(wms_exception_format, "BLANK") == 0 ||
      strcasecmp(wms_exception_format, "application/vnd.ogc.se_inimage")== 0 ||
      strcasecmp(wms_exception_format, "application/vnd.ogc.se_blank") == 0)
  {
    int blank = 0;

    if (strcasecmp(wms_exception_format, "BLANK") == 0 ||
        strcasecmp(wms_exception_format, "application/vnd.ogc.se_blank") == 0)
    {
        blank = 1;
    }

    msWriteErrorImage(map, NULL, blank);

  }
  else if (strcasecmp(wms_exception_format, "WMS_XML") == 0) // Only in V1.0.0
  {
    printf("Content-type: text/xml%c%c",10,10);
    printf("<WMTException version=\"1.0.0\">\n");
    msWriteErrorXML(stdout);
    printf("</WMTException>\n");
  }
  else // XML error, the default: SE_XML (1.0.1 to 1.0.7)
       // or application/vnd.ogc.se_xml (1.0.8, 1.1.0 and later)
  {
    if (wmtversion && strcasecmp(wmtversion, "1.0.7") <= 0)
    {
      // In V1.0.1 to 1.0.7, the MIME type was text/xml
      printf("Content-type: text/xml%c%c",10,10);

      msOWSPrintMetadata(stdout, map->web.metadata,
                         NULL, "wms_encoding", OWS_NOERR,
                "<?xml version='1.0' encoding=\"%s\" standalone=\"no\" ?>\n",
                    "ISO-8859-1");
      printf("<!DOCTYPE ServiceExceptionReport SYSTEM \"http://www.digitalearth.gov/wmt/xml/exception_1_0_1.dtd\">\n");

      printf("<ServiceExceptionReport version=\"1.0.1\">\n");
    }
    else
    {
      // In V1.0.8, 1.1.0 and later, we have OGC-specific MIME types
      // we cannot return anything else than application/vnd.ogc.se_xml here.
      printf("Content-type: application/vnd.ogc.se_xml%c%c",10,10);

      msOWSPrintMetadata(stdout, map->web.metadata,
                         NULL, "wms_encoding", OWS_NOERR,
                "<?xml version='1.0' encoding=\"%s\" standalone=\"no\" ?>\n",
                    "ISO-8859-1");
      printf("<!DOCTYPE ServiceExceptionReport SYSTEM \"http://schemas.opengis.net/wms/1.1.1/WMS_exception_1_1_1.dtd\">\n");

      printf("<ServiceExceptionReport version=\"1.1.0\">\n");
    }

    if (exception_code)
      printf("<ServiceException code=\"%s\">\n", exception_code);
    else
      printf("<ServiceException>\n");
    msWriteErrorXML(stdout);
    printf("</ServiceException>\n");
    printf("</ServiceExceptionReport>\n");
  }

  return MS_FAILURE; // so that we can call 'return msWMSException();' anywhere
}


/*
**
*/
int msWMSLoadGetMapParams(mapObj *map, const char *wmtver,
                          char **names, char **values, int numentries)
{
  int i, adjust_extent = MS_FALSE;
  int iUnits = -1;
  int nLayerOrder = 0;
  int transparent = MS_NOOVERRIDE;
  outputFormatObj *format = NULL;
  int validlayers = 0;
  char *styles = NULL;
  int numlayers = 0;
  char **layers = NULL;
  int layerfound =0;
  int invalidlayers = 0;

  // Some of the getMap parameters are actually required depending on the
  // request, but for now we assume all are optional and the map file
  // defaults will apply.

  for(i=0; map && i<numentries; i++)
  {
    // getMap parameters
    if (strcasecmp(names[i], "LAYERS") == 0)
    {
      int  j, k, iLayer;

      layers = split(values[i], ',', &numlayers);
      if (layers==NULL || numlayers < 1) {
        msSetError(MS_WMSERR, "At least one layer name required in LAYERS.",
                   "msWMSLoadGetMapParams()");
        return msWMSException(map, wmtver, NULL);
      }


      for (iLayer=0; iLayer < map->numlayers; iLayer++)
          map->layerorder[iLayer] = iLayer;

      for(j=0; j<map->numlayers; j++)
      {
        // Keep only layers with status=DEFAULT by default
        // Layer with status DEFAULT is drawn first.
        if (map->layers[j].status != MS_DEFAULT)
           map->layers[j].status = MS_OFF;
        else
           map->layerorder[nLayerOrder++] = j;
      }

      for (k=0; k<numlayers; k++)
      {
          layerfound = 0;
          for (j=0; j<map->numlayers; j++)
          {
              // Turn on selected layers only.
              if ((map->layers[j].name &&
                   strcasecmp(map->layers[j].name, layers[k]) == 0) ||
                  (map->name && strcasecmp(map->name, layers[k]) == 0) ||
                  (map->layers[j].group && strcasecmp(map->layers[j].group, layers[k]) == 0))
              {
                  map->layers[j].status = MS_ON;
                  map->layerorder[nLayerOrder++] = j;
                  validlayers++;
                  layerfound = 1;
              }
          }
          if (layerfound == 0)
            invalidlayers++;

      }

      // set all layers with status off at end of array
      for (j=0; j<map->numlayers; j++)
      {
         if (map->layers[j].status == MS_OFF)
           map->layerorder[nLayerOrder++] = j;
      }

      msFreeCharArray(layers, numlayers);
    }
    else if (strcasecmp(names[i], "STYLES") == 0) {
        styles = values[i];

    }
    else if (strcasecmp(names[i], "SRS") == 0) {
      // SRS is in format "EPSG:epsg_id" or "AUTO:proj_id,unit_id,lon0,lat0"
      if (strncasecmp(values[i], "EPSG:", 5) == 0)
      {
        // SRS=EPSG:xxxx

        char buffer[32];
        sprintf(buffer, "init=epsg:%.20s", values[i]+5);

        if (msLoadProjectionString(&(map->projection), buffer) != 0)
          return msWMSException(map, wmtver, NULL);

        iUnits = GetMapserverUnitUsingProj(&(map->projection));
        if (iUnits != -1)
          map->units = iUnits;
      }
      else if (strncasecmp(values[i], "AUTO:", 5) == 0)
      {
        // SRS=AUTO:proj_id,unit_id,lon0,lat0

        if (msLoadProjectionString(&(map->projection), values[i]) != 0)
          return msWMSException(map, wmtver, NULL);

        iUnits = GetMapserverUnitUsingProj(&(map->projection));
        if (iUnits != -1)
          map->units = iUnits;
      }
      else
      {
        msSetError(MS_WMSERR,
                   "Unsupported SRS namespace (only EPSG and AUTO currently supported).",
                   "msWMSLoadGetMapParams()");
        return msWMSException(map, wmtver, "InvalidSRS");
      }
    }
    else if (strcasecmp(names[i], "BBOX") == 0) {
      char **tokens;
      int n;
      tokens = split(values[i], ',', &n);
      if (tokens==NULL || n != 4) {
        msSetError(MS_WMSERR, "Wrong number of arguments for BBOX.",
                   "msWMSLoadGetMapParams()");
        return msWMSException(map, wmtver, NULL);
      }
      map->extent.minx = atof(tokens[0]);
      map->extent.miny = atof(tokens[1]);
      map->extent.maxx = atof(tokens[2]);
      map->extent.maxy = atof(tokens[3]);

      msFreeCharArray(tokens, n);

      //validate bbox values
      if ( map->extent.minx >= map->extent.maxx ||
           map->extent.miny >=  map->extent.maxy)
      {
          msSetError(MS_WMSERR,
                   "Invalid values for BBOX.",
                   "msWMSLoadGetMapParams()");
          return msWMSException(map, wmtver, NULL);
      }
      adjust_extent = MS_TRUE;
    }
    else if (strcasecmp(names[i], "WIDTH") == 0) {
      map->width = atoi(values[i]);
    }
    else if (strcasecmp(names[i], "HEIGHT") == 0) {
      map->height = atoi(values[i]);
    }
    else if (strcasecmp(names[i], "FORMAT") == 0) {

      format = msSelectOutputFormat( map, values[i] );

      if( format == NULL ) {
        msSetError(MS_IMGERR,
                   "Unsupported output format (%s).",
                   "msWMSLoadGetMapParams()",
                   values[i] );
        return msWMSException(map, wmtver, "InvalidFormat");
      }

      msFree( map->imagetype );
      map->imagetype = strdup(values[i]);
    }
    else if (strcasecmp(names[i], "TRANSPARENT") == 0) {
      transparent = (strcasecmp(values[i], "TRUE") == 0);
    }
    else if (strcasecmp(names[i], "BGCOLOR") == 0) {
      long c;
      c = strtol(values[i], NULL, 16);
      map->imagecolor.red = (c/0x10000)&0xff;
      map->imagecolor.green = (c/0x100)&0xff;
      map->imagecolor.blue = c&0xff;
    }
#ifdef USE_OGR
/* -------------------------------------------------------------------- */
/*      SLD support :                                                   */
/*        - check if the SLD parameter is there. it is supposed to      */
/*      refer a valid URL containing an SLD document.                   */
/*        - check the SLD_BODY parameter that should contain the SLD    */
/*      xml string.                                                     */
/* -------------------------------------------------------------------- */
    else if (strcasecmp(names[i], "SLD") == 0 &&
             values[i] && strlen(values[i]) > 0) {
          msSLDApplySLDURL(map, values[i], -1, NULL);
    }
    else if (strcasecmp(names[i], "SLD_BODY") == 0 &&
               values[i] && strlen(values[i]) > 0) {
          msSLDApplySLD(map, values[i], -1, NULL);
    }
#endif

  }


  /*
  ** Apply the selected output format (if one was selected), and override
  ** the transparency if needed.
  */

  if( format != NULL )
      msApplyOutputFormat( &(map->outputformat), format, transparent,
                           MS_NOOVERRIDE, MS_NOOVERRIDE );

  //validate all layers given. If an invalid layer is sent, return an exception.
  if (validlayers == 0 || invalidlayers > 0)
  {
      msSetError(MS_WMSERR, "Invalid layer(s) given in the LAYERS parameter.",
                 "msWMSLoadGetMapParams()");
      return msWMSException(map, wmtver, "LayerNotDefined");
  }
  /* Validate requested image size. */
  if(map->width > map->maxsize || map->height > map->maxsize ||
     map->width < 1 || map->height < 1)
  {
      msSetError(MS_WMSERR, "Image size out of range, WIDTH and HEIGHT must be between 1 and %d pixels.", "msWMSLoadGetMapParams()", map->maxsize);

      // Restore valid default values in case errors INIMAGE are used
      map->width = 400;
      map->height= 300;
      return msWMSException(map, wmtver, NULL);
  }

  /* Validate Styles :
  ** mapserv does not advertize any styles (the default styles are the
  ** one that are used. So we are expecting here to have empty values
  ** for the styles parameter (...&STYLES=&...) Or for multiple Styles/Layers,
  ** we could have ...&STYLES=,,,. If that is not the
  ** case, we generate an exception.
  */
  if(styles && strlen(styles) > 0)
  {
    int length = strlen(styles);
    int ilength = 0;
    for (ilength = 0; ilength < length; ilength++)
      if (styles[ilength] != ',')
      {
          msSetError(MS_WMSERR, "Invalid style (%s). Mapserver supports only default styles and is expecting an empty string for the STYLES : STYLES= or STYLES=,,,",
               "msWMSLoadGetMapParams()", values[i]);
          return msWMSException(map, wmtver, "StyleNotDefined");
      }
  }

  /*
  ** WMS extents are edge to edge while MapServer extents are center of
  ** pixel to center of pixel.  Here we try to adjust the WMS extents
  ** in by half a pixel.  We wait till here becaus we want to ensure we
  ** are doing this in terms of the correct WIDTH and HEIGHT.
  */
  if( adjust_extent )
  {
      double	dx, dy;

      dx = (map->extent.maxx - map->extent.minx) / map->width;
      map->extent.minx += dx*0.5;
      map->extent.maxx -= dx*0.5;

      dy = (map->extent.maxy - map->extent.miny) / map->height;
      map->extent.miny += dy*0.5;
      map->extent.maxy -= dy*0.5;
  }

  return MS_SUCCESS;
}



/*
**
*/
static void msWMSPrintRequestCap(const char *wmtver, const char *request,
                              const char *script_url, const char *formats, ...)
{
  va_list argp;
  const char *fmt;

  printf("    <%s>\n", request);

  // We expect to receive a NULL-terminated args list of formats
  va_start(argp, formats);
  fmt = formats;
  while(fmt != NULL)
  {
      printf("      <Format>%s</Format>\n", fmt);
      fmt = va_arg(argp, const char *);
  }
  va_end(argp);

  printf("      <DCPType>\n");
  printf("        <HTTP>\n");

  if (strcasecmp(wmtver, "1.0.0") == 0) {
    printf("          <Get onlineResource=\"%s\" />\n", script_url);
    printf("          <Post onlineResource=\"%s\" />\n", script_url);
  } else {
    printf("          <Get><OnlineResource xmlns:xlink=\"http://www.w3.org/1999/xlink\" xlink:href=\"%s\"/></Get>\n", script_url);
    printf("          <Post><OnlineResource xmlns:xlink=\"http://www.w3.org/1999/xlink\" xlink:href=\"%s\"/></Post>\n", script_url);
  }

  printf("        </HTTP>\n");
  printf("      </DCPType>\n");
  printf("    </%s>\n", request);
}



/*
** msWMSPrintScaleHint()
**
** Print a ScaleHint tag for this layer if applicable.
**
** (see WMS 1.1.0 sect. 7.1.5.4) The WMS defines the scalehint values as
** the ground distance in meters of the southwest to northeast diagonal of
** the central pixel of a map.  ScaleHint values are the min and max
** recommended values of that diagonal.
*/
void msWMSPrintScaleHint(const char *tabspace, double minscale,
                         double maxscale, double resolution)
{
  double scalehintmin=0.0, scalehintmax=0.0, diag;

  diag = sqrt(2.0);

  if (minscale > 0)
    scalehintmin = diag*(minscale/resolution)/msInchesPerUnit(MS_METERS,0);
  if (maxscale > 0)
    scalehintmax = diag*(maxscale/resolution)/msInchesPerUnit(MS_METERS,0);

  if (scalehintmin > 0.0 || scalehintmax > 0.0)
  {
      printf("%s<ScaleHint min=\"%g\" max=\"%g\" />\n",
             tabspace, scalehintmin, scalehintmax);
      if (scalehintmax == 0.0)
          printf("%s<!-- WARNING: Only MINSCALE and no MAXSCALE specified in "
                 "the mapfile. A default value of 0 has been returned for the "
                 "Max ScaleHint but this is probably not what you want. -->\n",
                 tabspace);
  }
}


/*
** msDumpLayer()
*/
int msDumpLayer(mapObj *map, layerObj *lp, const char* wmtver,
                const char *indent)
{
   rectObj ext;
   const char *value;
   char **tokens;
   int n, i;
   const char *projstring;

   if (strcasecmp(wmtver, "1.0.7") <= 0)
   {
       printf("%s    <Layer queryable=\"%d\">\n",
              indent, msIsLayerQueryable(lp));
   }
   else
   {
       // 1.0.8, 1.1.0 and later: opaque and cascaded are new.
       int cascaded=0, opaque=0;
       if ((value = msLookupHashTable(lp->metadata, "wms_opaque")) != NULL)
           opaque = atoi(value);
       if (lp->connectiontype == MS_WMS)
           cascaded = 1;

       printf("%s    <Layer queryable=\"%d\" opaque=\"%d\" cascaded=\"%d\">\n",
              indent, msIsLayerQueryable(lp), opaque, cascaded);
   }

   msOWSPrintParam(stdout, "LAYER.NAME", lp->name, OWS_WARN,
              "        <Name>%s</Name>\n", NULL);
   // the majority of this section is dependent on appropriately named metadata in the LAYER object
   msOWSPrintMetadata(stdout, lp->metadata, NULL, "wms_title", OWS_WARN,
                 "        <Title>%s</Title>\n", lp->name);

   msOWSPrintMetadata(stdout, lp->metadata, NULL, "wms_abstract", OWS_NOERR,
                 "        <Abstract>%s</Abstract>\n", NULL);

   if (strcasecmp(wmtver, "1.0.0") == 0)
   {
       // <Keywords> in V 1.0.0
       // The 1.0.0 spec doesn't specify which delimiter to use so let's use spaces
       msOWSPrintMetadataList(stdout, lp->metadata, NULL, "wms_keywordlist",
                              "        <Keywords>", "        </Keywords>\n",
                              "%s ", NULL);
   }
   else
   {
       // <KeywordList><Keyword> ... in V1.0.6+
       msOWSPrintMetadataList(stdout, lp->metadata, NULL, "wms_keywordlist",
                              "        <KeywordList>\n", "        </KeywordList>\n",
                              "          <Keyword>%s</Keyword>\n", NULL);
   }

   if (msGetEPSGProj(&(map->projection),map->web.metadata,MS_FALSE) == NULL)
   {
       if (strcasecmp(wmtver, "1.1.0") > 0)
       {
           projstring = msGetEPSGProj(&(lp->projection), lp->metadata,
                                      MS_FALSE);
           if (!projstring)
             fprintf(stdout, "<!-- WARNING: Mandatory mapfile parameter '%s' was missing in this context. -->\n", "(at least one of) MAP.PROJECTION, LAYER.PROJECTION or wms_srs metadata");
           else
           {
               tokens = split(projstring, ' ', &n);
               if (tokens && n > 0)
               {
                   for(i=0; i<n; i++)
                       fprintf(stdout, "        <SRS>%s</SRS>\n", tokens[i]);

                    msFreeCharArray(tokens, n);
               }
           }
       }
       else
         // If map has no proj then every layer MUST have one or produce a warning
         msOWSPrintParam(stdout, "(at least one of) MAP.PROJECTION, LAYER.PROJECTION or wms_srs metadata",
                         msGetEPSGProj(&(lp->projection), lp->metadata, MS_FALSE),
                         OWS_WARN, "        <SRS>%s</SRS>\n", NULL);
   }
   else
   {
       //starting 1.1.1 SRS are given in individual tags
       if (strcasecmp(wmtver, "1.1.0") > 0)
       {
           projstring = msGetEPSGProj(&(lp->projection), lp->metadata,
                                      MS_FALSE);
           if (projstring)
           {
               tokens = split(projstring, ' ', &n);
               if (tokens && n > 0)
               {
                   for(i=0; i<n; i++)
                       fprintf(stdout, "        <SRS>%s</SRS>\n", tokens[i]);

                    msFreeCharArray(tokens, n);
               }
           }
       }
       else
       // No warning required in this case since there's at least a map proj.
         msOWSPrintParam(stdout, " LAYER.PROJECTION (or wms_srs metadata)",
                         msGetEPSGProj(&(lp->projection), lp->metadata, MS_FALSE),
                         OWS_NOERR, "        <SRS>%s</SRS>\n", NULL);
   }

   // If layer has no proj set then use map->proj for bounding box.
   if (msOWSGetLayerExtent(map, lp, &ext) == MS_SUCCESS)
   {
       if(lp->projection.numargs > 0)
       {
           msOWSPrintLatLonBoundingBox(stdout, "        ", &(ext),
                                       &(lp->projection), OWS_WMS);
           msOWSPrintBoundingBox( stdout,"        ", &(ext), &(lp->projection),
                                  lp->metadata );
       }
       else
       {
           msOWSPrintLatLonBoundingBox(stdout, "        ", &(ext),
                                       &(map->projection), OWS_WMS);
           msOWSPrintBoundingBox(stdout,"        ", &(ext), &(map->projection),
                                  map->web.metadata );
       }
   }

   msWMSPrintScaleHint("        ", lp->minscale, lp->maxscale, map->resolution);

   printf("%s    </Layer>\n", indent);

   return MS_SUCCESS;
}



/*
** msWMSGetCapabilities()
*/
int msWMSGetCapabilities(mapObj *map, const char *wmtver, cgiRequestObj *req)
{
  const char *dtd_url = NULL;
  char *script_url=NULL, *script_url_encoded=NULL;
  char *pszMimeType=NULL;

  // Decide which version we're going to return.
  if (wmtver && strcasecmp(wmtver, "1.0.7") < 0) {
    wmtver = "1.0.0";
    dtd_url = "http://www.digitalearth.gov/wmt/xml/capabilities_1_0_0.dtd";
  }
  else if (wmtver && strcasecmp(wmtver, "1.0.8") < 0) {
    wmtver = "1.0.7";
    dtd_url = "http://www.digitalearth.gov/wmt/xml/capabilities_1_0_7.dtd";
  }
  else if (wmtver && strcasecmp(wmtver, "1.1.0") < 0) {
    wmtver = "1.0.8";
    dtd_url = "http://www.digitalearth.gov/wmt/xml/capabilities_1_0_8.dtd";
  }
  else if (wmtver && strcasecmp(wmtver, "1.1.0") == 0) {
    wmtver = "1.1.0";
    dtd_url = "http://www.digitalearth.gov/wmt/xml/capabilities_1_1_0.dtd";
  }
  else {
    wmtver = "1.1.1";
    dtd_url = "http://schemas.opengis.net/wms/1.1.1/WMS_MS_Capabilities.dtd";
  }

  // We need this server's onlineresource.
  // Default to use the value of the "onlineresource" metadata, and if not
  // set then build it: "http://$(SERVER_NAME):$(SERVER_PORT)$(SCRIPT_NAME)?"
  // the returned string should be freed once we're done with it.
  if ((script_url=msOWSGetOnlineResource(map, "wms_onlineresource", req)) == NULL ||
      (script_url_encoded = msEncodeHTMLEntities(script_url)) == NULL)
  {
      return msWMSException(map, wmtver, NULL);
  }

  if (strcasecmp(wmtver, "1.0.7") <= 0)
      printf("Content-type: text/xml%c%c",10,10);  // 1.0.0 to 1.0.7
  else
      printf("Content-type: application/vnd.ogc.wms_xml%c%c",10,10);  // 1.0.8, 1.1.0 and later

  msOWSPrintMetadata(stdout, map->web.metadata,
                     NULL, "wms_encoding", OWS_NOERR,
                "<?xml version='1.0' encoding=\"%s\" standalone=\"no\" ?>\n",
                "ISO-8859-1");
  printf("<!DOCTYPE WMT_MS_Capabilities SYSTEM \"%s\"\n", dtd_url);
  printf(" [\n");

  // some mapserver specific declarations will go here
  printf(" <!ELEMENT VendorSpecificCapabilities EMPTY>\n");

  printf(" ]>  <!-- end of DOCTYPE declaration -->\n\n");

  printf("<WMT_MS_Capabilities version=\"%s\">\n",wmtver);

  // Report MapServer Version Information
  printf("\n<!-- %s -->\n\n", msGetVersion());

  // WMS definition
  printf("<Service>\n");

  // Service name is defined by the spec and changed at v1.0.0
  if (strcasecmp(wmtver, "1.0.7") <= 0)
      printf("  <Name>GetMap</Name>\n");  // v 1.0.0 to 1.0.7
  else
      printf("  <Name>OGC:WMS</Name>\n"); // v 1.1.0+

  // the majority of this section is dependent on appropriately named metadata in the WEB object
  msOWSPrintMetadata(stdout, map->web.metadata, NULL, "wms_title", OWS_WARN,
                "  <Title>%s</Title>\n", map->name);
  msOWSPrintMetadata(stdout, map->web.metadata, NULL, "wms_abstract", OWS_NOERR,
                "  <Abstract>%s</Abstract>\n", NULL);

  if (strcasecmp(wmtver, "1.0.0") == 0)
  {
      // <Keywords> in V 1.0.0
      // The 1.0.0 spec doesn't specify which delimiter to use so let's use spaces
      msOWSPrintMetadataList(stdout, map->web.metadata,
                             NULL, "wms_keywordlist",
                             "        <Keywords>", "        </Keywords>\n",
                             "%s ", NULL);
  }
  else
  {
      // <KeywordList><Keyword> ... in V1.0.6+
      msOWSPrintMetadataList(stdout, map->web.metadata,
                             NULL, "wms_keywordlist",
                             "        <KeywordList>\n", "        </KeywordList>\n",
                             "          <Keyword>%s</Keyword>\n", NULL);
  }

  if (strcasecmp(wmtver, "1.0.0") == 0)
    printf("  <OnlineResource>%s</OnlineResource>\n", script_url_encoded);
  else
    printf("  <OnlineResource xmlns:xlink=\"http://www.w3.org/1999/xlink\" xlink:href=\"%s\"/>\n", script_url_encoded);

  // contact information is a required element in 1.0.7 but the
  // sub-elements such as ContactPersonPrimary, etc. are not!
  // In 1.1.0, ContactInformation becomes optional.
  msOWSPrintContactInfo(stdout, "  ", wmtver, map->web.metadata);

  msOWSPrintMetadata(stdout, map->web.metadata, NULL, "wms_fees", OWS_NOERR,
                "  <Fees>%s</Fees>\n", NULL);

  msOWSPrintMetadata(stdout, map->web.metadata, NULL, "wms_accessconstraints",
             OWS_NOERR, "  <AccessConstraints>%s</AccessConstraints>\n", NULL);

  printf("</Service>\n\n");

  // WMS capabilities definitions
  printf("<Capability>\n");
  printf("  <Request>\n");

  if (strcasecmp(wmtver, "1.0.7") <= 0)
  {
    // WMS 1.0.0 to 1.0.7 - We don't try to use outputformats list here for now
    msWMSPrintRequestCap(wmtver, "Map", script_url_encoded, ""
#ifdef USE_GD_GIF
                      "<GIF />"
#endif
#ifdef USE_GD_PNG
                      "<PNG />"
#endif
#ifdef USE_GD_JPEG
                      "<JPEG />"
#endif
#ifdef USE_GD_WBMP
                      "<WBMP />"
#endif
                      , NULL);
    msWMSPrintRequestCap(wmtver, "Capabilities", script_url_encoded, "<WMS_XML />", NULL);
    msWMSPrintRequestCap(wmtver, "FeatureInfo", script_url_encoded, "<MIME /><GML.1 />", NULL);
  }
  else
  {
    char *mime_list[20];
    // WMS 1.0.8, 1.1.0 and later
    // Note changes to the request names, their ordering, and to the formats

    msWMSPrintRequestCap(wmtver, "GetCapabilities", script_url_encoded,
                    "application/vnd.ogc.wms_xml",
                    NULL);

    msGetOutputFormatMimeList(map,mime_list,sizeof(mime_list)/sizeof(char*));
    msWMSPrintRequestCap(wmtver, "GetMap", script_url_encoded,
                    mime_list[0], mime_list[1], mime_list[2], mime_list[3],
                    mime_list[4], mime_list[5], mime_list[6], mime_list[7],
                    mime_list[8], mime_list[9], mime_list[10], mime_list[11],
                    mime_list[12], mime_list[13], mime_list[14], mime_list[15],
                    mime_list[16], mime_list[17], mime_list[18], mime_list[19],
                    NULL );

    pszMimeType = msLookupHashTable(map->web.metadata, "WMS_FEATURE_INFO_MIME_TYPE");
    if (pszMimeType && strcasecmp(pszMimeType, "NONE") == 0)
       msWMSPrintRequestCap(wmtver, "GetFeatureInfo", script_url_encoded,
                       "text/plain",
                       "application/vnd.ogc.gml",
                       NULL);
    else
    if (pszMimeType)
       msWMSPrintRequestCap(wmtver, "GetFeatureInfo", script_url_encoded,
                       "text/plain",
                       pszMimeType,
                       "application/vnd.ogc.gml",
                       NULL);
    else
       msWMSPrintRequestCap(wmtver, "GetFeatureInfo", script_url_encoded,
                       "text/plain",
                       "text/html",
                       "application/vnd.ogc.gml",
                       NULL);

    msWMSPrintRequestCap(wmtver, "DescribeLayer", script_url_encoded,
                         "text/xml",
                         NULL);

    msWMSPrintRequestCap(wmtver, "GetLegendGraphic", script_url_encoded,
                    mime_list[0], mime_list[1], mime_list[2], mime_list[3],
                    mime_list[4], mime_list[5], mime_list[6], mime_list[7],
                    mime_list[8], mime_list[9], mime_list[10], mime_list[11],
                    mime_list[12], mime_list[13], mime_list[14], mime_list[15],
                    mime_list[16], mime_list[17], mime_list[18], mime_list[19],
                    NULL );
  }

  printf("  </Request>\n");

  printf("  <Exception>\n");
  if (strcasecmp(wmtver, "1.0.7") <= 0)
      printf("    <Format><BLANK /><INIMAGE /><WMS_XML /></Format>\n");
  else
  {
      // 1.0.8, 1.1.0 and later
      printf("    <Format>application/vnd.ogc.se_xml</Format>\n");
      printf("    <Format>application/vnd.ogc.se_inimage</Format>\n");
      printf("    <Format>application/vnd.ogc.se_blank</Format>\n");
  }
  printf("  </Exception>\n");


  printf("  <VendorSpecificCapabilities />\n"); // nothing yet

  //SLD support
  printf("  <UserDefinedSymbolization SupportSLD=\"1\" UserLayer=\"0\" UserStyle=\"1\" RemoteWFS=\"0\"/>\n");

  // Top-level layer with map extents and SRS, encloses all map layers
  printf("  <Layer>\n");

  // Layer Name is optional but title is mandatory.
  msOWSPrintParam(stdout, "MAP.NAME", map->name, OWS_NOERR,
                  "    <Name>%s</Name>\n", NULL);
  msOWSPrintMetadata(stdout, map->web.metadata, NULL, "wms_title", OWS_WARN,
                "    <Title>%s</Title>\n", map->name);

  // According to normative comments in the 1.0.7 DTD, the root layer's SRS tag
  // is REQUIRED.  It also suggests that we use an empty SRS element if there
  // is no common SRS.
  msOWSPrintParam(stdout, "MAP.PROJECTION (or wms_srs metadata)",
             msGetEPSGProj(&(map->projection), map->web.metadata, MS_FALSE),
             OWS_WARN, "    <SRS>%s</SRS>\n", "");

  msOWSPrintLatLonBoundingBox(stdout, "    ", &(map->extent),
                              &(map->projection), OWS_WMS);
  msOWSPrintBoundingBox( stdout, "    ", &(map->extent), &(map->projection),
                         map->web.metadata );
  msWMSPrintScaleHint("    ", map->web.minscale, map->web.maxscale,
                      map->resolution);


  //
  // Dump list of layers organized by groups.  Layers with no group are listed
  // individually, at the same level as the groups in the layer hierarchy
  //
  if (map->numlayers)
  {
     int i, j;
     char *pabLayerProcessed = NULL;

     // We'll use this array of booleans to track which layer/group have been
     // processed already
     pabLayerProcessed = (char *)calloc(map->numlayers, sizeof(char*));

     for(i=0; i<map->numlayers; i++)
     {
         layerObj *lp;
         lp = &(map->layers[i]);

         if (pabLayerProcessed[i])
             continue;  // Layer has already been handled

         if (lp->group == NULL || strlen(lp->group) == 0)
         {
             // This layer is not part of a group... dump it directly
             msDumpLayer(map, lp, wmtver, "");
             pabLayerProcessed[i] = 1;
         }
         else
         {
             // Beginning of a new group... enclose the group in a layer block
             printf("    <Layer>\n");

             // Layer Name is optional but title is mandatory.
             msOWSPrintParam(stdout, "GROUP.NAME", lp->group, OWS_NOERR,
                        "      <Name>%s</Name>\n", NULL);
             msOWSPrintGroupMetadata(stdout, map, lp->group,
                                     NULL, "WMS_GROUP_TITLE", OWS_WARN,
                                "      <Title>%s</Title>\n", lp->group);

             // Dump all layers for this group
             for(j=i; j<map->numlayers; j++)
             {
                 if (!pabLayerProcessed[j] &&
                     map->layers[j].group &&
                     strcmp(lp->group, map->layers[j].group) == 0 )
                 {
                     msDumpLayer(map, &(map->layers[j]), wmtver, "  ");
                     pabLayerProcessed[j] = 1;
                 }
             }

             // Close group layer block
             printf("    </Layer>\n");
         }
     }

     free(pabLayerProcessed);
  }

  printf("  </Layer>\n");

  printf("</Capability>\n");
  printf("</WMT_MS_Capabilities>\n");

  free(script_url);
  free(script_url_encoded);

  return(MS_SUCCESS);
}

/*
 * This function look for params that can be used
 * by mapserv when generating template.
*/
int msTranslateWMS2Mapserv(char **names, char **values, int *numentries)
{
   int i=0;
   int tmpNumentries = *numentries;;

   for (i=0; i<*numentries; i++)
   {
      if (strcasecmp("X", names[i]) == 0)
      {
         values[tmpNumentries] = strdup(values[i]);
         names[tmpNumentries] = strdup("img.x");

         tmpNumentries++;
      }
      else
      if (strcasecmp("Y", names[i]) == 0)
      {
         values[tmpNumentries] = strdup(values[i]);
         names[tmpNumentries] = strdup("img.y");

         tmpNumentries++;
      }
      else
      if (strcasecmp("LAYERS", names[i]) == 0)
      {
         char **layers;
         int tok;
         int j;

         layers = split(values[i], ',', &tok);

         for (j=0; j<tok; j++)
         {
            values[tmpNumentries] = layers[j];
            layers[j] = NULL;
            names[tmpNumentries] = strdup("layer");

            tmpNumentries++;
         }

         free(layers);
      }
      else
      if (strcasecmp("QUERY_LAYERS", names[i]) == 0)
      {
         char **layers;
         int tok;
         int j;

         layers = split(values[i], ',', &tok);

         for (j=0; j<tok; j++)
         {
            values[tmpNumentries] = layers[j];
            layers[j] = NULL;
            names[tmpNumentries] = strdup("qlayer");

            tmpNumentries++;
         }

         free(layers);
      }
      else
      if (strcasecmp("BBOX", names[i]) == 0)
      {
         char *imgext;

         // Note gsub() works on the string itself, so we need to make a copy
         imgext = strdup(values[i]);
         imgext = gsub(imgext, ",", " ");

         values[tmpNumentries] = imgext;
         names[tmpNumentries] = strdup("imgext");

         tmpNumentries++;
      }
   }

   *numentries = tmpNumentries;

   return MS_SUCCESS;
}

/*
** msWMSGetMap()
*/
int msWMSGetMap(mapObj *map, const char *wmtver, char **names, char **values, int numentries)
{
  imageObj *img;

  // __TODO__ msDrawMap() will try to adjust the extent of the map
  // to match the width/height image ratio.
  // The spec states that this should not happen so that we can deliver
  // maps to devices with non-square pixels.


  img = msDrawMap(map);
  if (img == NULL)
      return msWMSException(map, wmtver, NULL);

  printf("Content-type: %s%c%c", MS_IMAGE_MIME_TYPE(map->outputformat), 10,10);
  if (msSaveImage(map, img, NULL) != MS_SUCCESS)
      return msWMSException(map, wmtver, NULL);

  msFreeImage(img);

  return(MS_SUCCESS);
}

int msDumpResult(mapObj *map, int bFormatHtml, const char* wmtver, int feature_count)
{
   int numresults=0;
   int i;

   for(i=0; i<map->numlayers && numresults<feature_count; i++)
   {
      int j, k;
      layerObj *lp;
      lp = &(map->layers[i]);

      if(lp->status != MS_ON || lp->resultcache==NULL || lp->resultcache->numresults == 0)
        continue;

      if(msLayerOpen(lp) != MS_SUCCESS || msLayerGetItems(lp) != MS_SUCCESS)
        return msWMSException(map, wmtver, NULL);

      printf("\nLayer '%s'\n", lp->name);

      for(j=0; j<lp->resultcache->numresults && numresults<feature_count; j++) {
        shapeObj shape;

        msInitShape(&shape);
        if (msLayerGetShape(lp, &shape, lp->resultcache->results[j].tileindex, lp->resultcache->results[j].shapeindex) != MS_SUCCESS)
          return msWMSException(map, wmtver, NULL);

        printf("  Feature %ld: \n", lp->resultcache->results[j].shapeindex);

        for(k=0; k<lp->numitems; k++)
	  printf("    %s = '%s'\n", lp->items[k], shape.values[k]);

        msFreeShape(&shape);
        numresults++;
      }

      msLayerClose(lp);
    }

   return numresults;
}


/*
** msWMSFeatureInfo()
*/
int msWMSFeatureInfo(mapObj *map, const char *wmtver, char **names, char **values, int numentries)
{
  int i, feature_count=1, numlayers_found=0;
  pointObj point = {-1.0, -1.0};
  const char *info_format="MIME";
  double cellx, celly;
  errorObj *ms_error = msGetErrorObj();
  int status;
  const char *pszMimeType=NULL;

  pszMimeType = msLookupHashTable(map->web.metadata, "WMS_FEATURE_INFO_MIME_TYPE");
  if (!pszMimeType)
  {
     pszMimeType = "text/html";
  }


  for(i=0; map && i<numentries; i++) {
    if(strcasecmp(names[i], "QUERY_LAYERS") == 0) {
      char **layers;
      int numlayers, j, k;

      layers = split(values[i], ',', &numlayers);
      if(layers==NULL || numlayers < 1) {
        msSetError(MS_WMSERR, "At least one layer name required in QUERY_LAYERS.", "msWMSFeatureInfo()");
        return msWMSException(map, wmtver, NULL);
      }

      for(j=0; j<map->numlayers; j++) {
        // Force all layers OFF by default
	map->layers[j].status = MS_OFF;

        for(k=0; k<numlayers; k++) {
          if ((map->layers[j].name && strcasecmp(map->layers[j].name, layers[k]) == 0) ||
              (map->name && strcasecmp(map->name, layers[k]) == 0) ||
              (map->layers[j].group && strcasecmp(map->layers[j].group, layers[k]) == 0))
            {
              map->layers[j].status = MS_ON;
              numlayers_found++;
            }
        }
      }

      msFreeCharArray(layers, numlayers);
    } else if (strcasecmp(names[i], "INFO_FORMAT") == 0)
      info_format = values[i];
    else if (strcasecmp(names[i], "FEATURE_COUNT") == 0)
      feature_count = atoi(values[i]);
    else if(strcasecmp(names[i], "X") == 0)
      point.x = atof(values[i]);
    else if (strcasecmp(names[i], "Y") == 0)
      point.y = atof(values[i]);
    else if (strcasecmp(names[i], "RADIUS") == 0)
    {
        // RADIUS in pixels.
        // This is not part of the spec, but some servers such as cubeserv
        // support it as a vendor-specific feature.
        // It's easy for MapServer to handle this so let's do it!
        int j;
        for(j=0; j<map->numlayers; j++)
        {
            map->layers[j].tolerance = atoi(values[i]);
            map->layers[j].toleranceunits = MS_PIXELS;
        }
    }

  }

  if(numlayers_found == 0) {
    msSetError(MS_WMSERR, "Required QUERY_LAYERS parameter missing for getFeatureInfo.", "msWMSFeatureInfo()");
    return msWMSException(map, wmtver, "LayerNotQueryable");
  }

/* -------------------------------------------------------------------- */
/*      check if all layers selected are queryable. If not send an      */
/*      exception.                                                      */
/* -------------------------------------------------------------------- */
  for (i=0; i<map->numlayers; i++)
  {
      if (map->layers[i].status == MS_ON && !msIsLayerQueryable(&map->layers[i]))
      {
          msSetError(MS_WMSERR, "Requested layer(s) are not queryable.", "msWMSFeatureInfo()");
          return msWMSException(map, wmtver, "LayerNotQueryable");
      }
  }
  if(point.x == -1.0 || point.y == -1.0) {
    msSetError(MS_WMSERR, "Required X/Y parameters missing for getFeatureInfo.", "msWMSFeatureInfo()");
    return msWMSException(map, wmtver, NULL);
  }

  // Perform the actual query
  cellx = MS_CELLSIZE(map->extent.minx, map->extent.maxx, map->width); // note: don't adjust extent, WMS assumes incoming extent is correct
  celly = MS_CELLSIZE(map->extent.miny, map->extent.maxy, map->height);
  point.x = MS_IMAGE2MAP_X(point.x, map->extent.minx, cellx);
  point.y = MS_IMAGE2MAP_Y(point.y, map->extent.maxy, celly);

  if(msQueryByPoint(map, -1, (feature_count==1?MS_SINGLE:MS_MULTIPLE), point, 0) != MS_SUCCESS)
    if(ms_error->code != MS_NOTFOUND) return msWMSException(map, wmtver, NULL);

  // Generate response
  if (strcasecmp(info_format, "MIME") == 0 ||
      strcasecmp(info_format, "text/plain") == 0) {

    // MIME response... we're free to use any valid MIME type
    int numresults = 0;

    printf("Content-type: text/plain%c%c", 10,10);
    printf("GetFeatureInfo results:\n");

    numresults = msDumpResult(map, 0, wmtver, feature_count);

    if (numresults == 0) printf("\n  Search returned no results.\n");

  } else if (strncasecmp(info_format, "GML", 3) == 0 ||  // accept GML.1 or GML
             strcasecmp(info_format, "application/vnd.ogc.gml") == 0) {

    if (strcasecmp(wmtver, "1.0.7") <= 0)
        printf("Content-type: text/xml%c%c", 10,10);
    else
        printf("Content-type: application/vnd.ogc.gml%c%c", 10,10);

    msGMLWriteQuery(map, NULL); // default is stdout

  } else
  if (pszMimeType && (strcmp(pszMimeType, info_format) == 0))
  {
     mapservObj *msObj;

     msObj = msAllocMapServObj();

     // Translate some vars from WMS to mapserv
     msTranslateWMS2Mapserv(names, values, &numentries);

     msObj->Map = map;
     msObj->request->ParamNames = names;
     msObj->request->ParamValues = values;
     msObj->Mode = QUERY;
     sprintf(msObj->Id, "%ld%d",(long)time(NULL),(int)getpid()); // asign now so it can be overridden
     msObj->request->NumParams = numentries;
     msObj->MapPnt.x = point.x;
     msObj->MapPnt.y = point.y;

     if ((status = msReturnTemplateQuery(msObj, (char*)pszMimeType,NULL)) != MS_SUCCESS)
         return msWMSException(map, wmtver, NULL);

     // We don't want to free the map, and param names/values since they
     // belong to the caller, set them to NULL before freeing the mapservObj
     msObj->Map = NULL;
     msObj->request->ParamNames = NULL;
     msObj->request->ParamValues = NULL;
     msObj->request->NumParams = 0;

     msFreeMapServObj(msObj);
  }
  else
  {
     msSetError(MS_WMSERR, "Unsupported INFO_FORMAT value (%s).", "msWMSFeatureInfo()", info_format);
     return msWMSException(map, wmtver, NULL);
  }

  return(MS_SUCCESS);
}

/*
** msWMSDescribeLayer()
*/
int msWMSDescribeLayer(mapObj *map, const char *wmtver, char **names,
                       char **values, int numentries)
{
  int i = 0;
  char **layers = NULL;
  int numlayers = 0;
  int j, k;
  layerObj *lp = NULL;
  const char *pszOnlineResMap = NULL, *pszOnlineResLyr = NULL;

   for(i=0; map && i<numentries; i++) {
     if(strcasecmp(names[i], "LAYERS") == 0) {

      layers = split(values[i], ',', &numlayers);
     }
   }

   msOWSPrintMetadata(stdout, map->web.metadata,
                      NULL, "wms_encoding", OWS_NOERR,
                      "<?xml version='1.0' encoding=\"%s\"?>\n",
                      "ISO-8859-1");
   printf("<!DOCTYPE WMS_DescribeLayerResponse>\n");
   printf("<WMS_DescribeLayerResponse version=\"1.1.0\" >\n");

   //check if map-level metadata wfs_onlineresource is available
   pszOnlineResMap = msLookupHashTable(map->web.metadata,"wfs_onlineresource");
   if (pszOnlineResMap && strlen(pszOnlineResMap) == 0)
       pszOnlineResMap = NULL;

   for(j=0; j<numlayers; j++)
   {
       for(k=0; k<map->numlayers; k++)
       {
         lp = &map->layers[k];
         if (lp->name && strcasecmp(lp->name, layers[j]) == 0)
         {
             /* Look for a WFS onlineresouce at the layer level and then at
              * the map level.
              */
           pszOnlineResLyr = msLookupHashTable(lp->metadata,
                                               "wfs_onlineresource");
           if (pszOnlineResLyr == NULL || strlen(pszOnlineResLyr) == 0)
               pszOnlineResLyr = pszOnlineResMap;

           if (pszOnlineResLyr && (lp->type == MS_LAYER_POINT ||
                                   lp->type == MS_LAYER_LINE ||
                                   lp->type == MS_LAYER_POLYGON ) )
           {
             char *pszOnlineResEncoded;
             pszOnlineResEncoded = msEncodeHTMLEntities(pszOnlineResLyr);

             printf("<LayerDescription name=\"%s\" wfs=\"%s\">\n",
                    lp->name, pszOnlineResEncoded);
             printf("<Query typeName=\"%s\" />\n", lp->name);
             printf("</LayerDescription>\n");

             msFree(pszOnlineResEncoded);
           }
           else
           {
             printf("<LayerDescription name=\"%s\"></LayerDescription>\n",
                    lp->name);
           }
           break;
         }
       }
   }

   printf("</WMS_DescribeLayerResponse>\n");

   if (layers)
     msFreeCharArray(layers, numlayers);

   return(MS_SUCCESS);
}


/*
** msWMSGetLegendGraphic()
*/
int msWMSGetLegendGraphic(mapObj *map, const char *wmtver, char **names,
                       char **values, int numentries)
{
    char *pszLayer = NULL;
    char *pszFormat = NULL;
	char *psRule = NULL;
    int iLayerIndex = -1;
    outputFormatObj *psFormat = NULL;
    imageObj *img;
    int i = 0;
    int nWidth = -1, nHeight =-1;

     for(i=0; map && i<numentries; i++)
     {
         if (strcasecmp(names[i], "LAYER") == 0)
         {
             pszLayer = values[i];
         }
         else if (strcasecmp(names[i], "WIDTH") == 0)
           nWidth = atoi(values[i]);
         else if (strcasecmp(names[i], "HEIGHT") == 0)
           nHeight = atoi(values[i]);
         else if (strcasecmp(names[i], "FORMAT") == 0)
           pszFormat = values[i];
#ifdef USE_OGR
/* -------------------------------------------------------------------- */
/*      SLD support :                                                   */
/*        - check if the SLD parameter is there. it is supposed to      */
/*      refer a valid URL containing an SLD document.                   */
/*        - check the SLD_BODY parameter that should contain the SLD    */
/*      xml string.                                                     */
/* -------------------------------------------------------------------- */
         else if (strcasecmp(names[i], "SLD") == 0 &&
                  values[i] && strlen(values[i]) > 0)
             msSLDApplySLDURL(map, values[i], -1, NULL);
         else if (strcasecmp(names[i], "SLD_BODY") == 0 &&
                  values[i] && strlen(values[i]) > 0)
             msSLDApplySLD(map, values[i], -1, NULL);
				 else if (strcasecmp(names[i], "RULE") == 0)
				 {
             psRule = values[i];
				 }

#endif
     }

     if (!pszLayer)
     {
         msSetError(MS_WMSERR, "Mandatory LAYER parameter missing in GetLegendGraphic request.", "msWMSGetLegendGraphic()");
         return msWMSException(map, wmtver, "LayerNotDefined");
     }
     if (!pszFormat)
     {
         msSetError(MS_WMSERR, "Mandatory FORMAT parameter missing in GetLegendGraphic request.", "msWMSGetLegendGraphic()");
         return msWMSException(map, wmtver, "InvalidFormat");
     }

     //check if layer name is valid. We only test the layer name and not
     //the group name.
     for (i=0; i<map->numlayers; i++)
     {
         if (map->layers[i].name &&
             strcasecmp(map->layers[i].name, pszLayer) == 0)
         {
             iLayerIndex = i;
             break;
         }
     }

     if (iLayerIndex == -1)
     {
         msSetError(MS_WMSERR, "Invalid layer given in the LAYER parameter.",
                 "msWMSGetLegendGraphic()");
         return msWMSException(map, wmtver, "LayerNotDefined");
     }

     //validate format
     psFormat = msSelectOutputFormat( map, pszFormat);
     if( psFormat == NULL )
     {
         msSetError(MS_IMGERR,
                    "Unsupported output format (%s).",
                    "msWMSGetLegendGraphic()",
                    pszFormat);
         return msWMSException(map, wmtver, "InvalidFormat");
     }

	 if ( psRule == NULL )
	 {
		 // turn off all other layers
		 for (i=0; i<map->numlayers; i++)
		 {
				 if (map->layers[i].name &&
						 strcasecmp(map->layers[i].name, pszLayer) != 0)
				 {
						 map->layers[i].status = MS_OFF;
				 }
		 }
		 img = msDrawLegend(map);
	 }
	 else
	 {
		 //set the map legend parameters
		 if (nWidth < 0)
		 {
				 if (map->legend.keysizex > 0)
					 nWidth = map->legend.keysizex;
				 else
					 nWidth = 20; //default values : this in not defined in the specs
		 }
		 if (nHeight < 0)
		 {
				 if (map->legend.keysizey > 0)
					 nHeight = map->legend.keysizey;
				 else
					 nHeight = 20;
		 }

		 img = msCreateLegendIcon(map, &map->layers[iLayerIndex], NULL,
															nWidth, nHeight);
	 }

     if (img == NULL)
      return msWMSException(map, wmtver, NULL);

     printf("Content-type: %s%c%c", MS_IMAGE_MIME_TYPE(map->outputformat), 10,10);
     if (msSaveImage(map, img, NULL) != MS_SUCCESS)
       return msWMSException(map, wmtver, NULL);

     msFreeImage(img);

     return(MS_SUCCESS);
}


/*
** msWMSGetStyles() : return an SLD document for all layers that
** have a status set to on or default.
*/
int msWMSGetStyles(mapObj *map, const char *wmtver, char **names,
                       char **values, int numentries)

{
    int i,j,k;
    int validlayer = 0;
    int numlayers = 0;
    char **layers = NULL;
    char  *sld = NULL;

    for(i=0; map && i<numentries; i++)
    {
        // getMap parameters
        if (strcasecmp(names[i], "LAYERS") == 0)
        {
            layers = split(values[i], ',', &numlayers);
            if (layers==NULL || numlayers < 1) {
                msSetError(MS_WMSERR, "At least one layer name required in LAYERS.",
                   "msWMSGetStyles()");
                return msWMSException(map, wmtver, NULL);
            }
            for(j=0; j<map->numlayers; j++)
               map->layers[j].status = MS_OFF;

            for (k=0; k<numlayers; k++)
            {
                for (j=0; j<map->numlayers; j++)
                {
                    if (map->layers[j].name &&
                        strcasecmp(map->layers[j].name, layers[k]) == 0)
                    {
                        map->layers[j].status = MS_ON;
                        validlayer =1;
                    }
                }
            }

            msFreeCharArray(layers, numlayers);
        }

    }

    //validate all layers given. If an invalid layer is sent, return an exception.
    if (validlayer == 0)
    {
        msSetError(MS_WMSERR, "Invalid layer(s) given in the LAYERS parameter.",
                   "msWMSGetStyles()");
        return msWMSException(map, wmtver, "LayerNotDefined");
    }

    printf("Content-type: application/vnd.ogc.sld+xml%c%c",10,10);
    sld = msSLDGenerateSLD(map, -1);
    if (sld)
    {
        printf("%s\n", sld);
        free(sld);
    }

    return(MS_SUCCESS);
}


#endif /* USE_WMS_SVR */


/*
** msWMSDispatch() is the entry point for WMS requests.
** - If this is a valid request then it is processed and MS_SUCCESS is returned
**   on success, or MS_FAILURE on failure.
** - If this does not appear to be a valid WMS request then MS_DONE
**   is returned and MapServer is expected to process this as a regular
**   MapServer request.
*/
int msWMSDispatch(mapObj *map, cgiRequestObj *req)
{
#ifdef USE_WMS_SVR
  int i, status;
  static char *wmtver = NULL, *request=NULL, *service=NULL, *format=NULL;

  /*
  ** Process Params common to all requests
  */
  // VERSION (WMTVER in 1.0.0) and REQUEST must be present in a valid request
  for(i=0; i<req->NumParams; i++) {
      if(strcasecmp(req->ParamNames[i], "VERSION") == 0)
        wmtver = req->ParamValues[i];
      if (strcasecmp(req->ParamNames[i], "WMTVER") == 0 && wmtver == NULL)
        wmtver = req->ParamValues[i];
      else if (strcasecmp(req->ParamNames[i], "REQUEST") == 0)
        request = req->ParamValues[i];
      else if (strcasecmp(req->ParamNames[i], "EXCEPTIONS") == 0)
        wms_exception_format = req->ParamValues[i];
      else if (strcasecmp(req->ParamNames[i], "SERVICE") == 0)
        service = req->ParamValues[i];
      else if (strcasecmp(req->ParamNames[i], "FORMAT") == 0)
        format = req->ParamValues[i];
  }

  /* If SERVICE is specified then it MUST be "WMS" */
  if (service != NULL && strcasecmp(service, "WMS") != 0)
      return MS_DONE;  /* Not a WMS request */

  /*
  ** Dispatch request... we should probably do some validation on VERSION here
  ** vs the versions we actually support.
  */
  if (request && (strcasecmp(request, "capabilities") == 0 ||
                  strcasecmp(request, "GetCapabilities") == 0) )
  {
      if (!wmtver)
          wmtver = "1.1.1";  // VERSION is optional with getCapabilities only
      if ((status = msOWSMakeAllLayersUnique(map)) != MS_SUCCESS)
          return msWMSException(map, wmtver, NULL);
      return msWMSGetCapabilities(map, wmtver, req);
  }
  else if (request && (strcasecmp(request, "context") == 0 ||
                       strcasecmp(request, "GetContext") == 0) )
  {
      /* Return a context document with all layers in this mapfile
       * This is not a standard WMS request.
       * __TODO__ The real implementation should actually return only context
       * info for selected layers in the LAYERS parameter.
       */
      const char *getcontext_enabled;
      getcontext_enabled = msLookupHashTable(map->web.metadata,
                                             "wms_getcontext_enabled");

      if (wmtver)
      {
          // VERSION, if specified, is Map Context version, not WMS version
          // Pass it via wms_context_version metadata
          msInsertHashTable(map->web.metadata, "wms_context_version", wmtver);
      }
      // Now set wmtver to 1.1.1 for error handling purposes
      wmtver = "1.1.1";

      if (getcontext_enabled==NULL || atoi(getcontext_enabled) == 0)
      {
        msSetError(MS_WMSERR, "GetContext not enabled on this server.",
                   "msWMSDispatch()");
        return msWMSException(map, wmtver, NULL);
      }

      if ((status = msOWSMakeAllLayersUnique(map)) != MS_SUCCESS)
          return msWMSException(map, wmtver, NULL);
      printf("Content-type: text/xml\n\n");
      if ( msWriteMapContext(map, stdout) != MS_SUCCESS )
          return msWMSException(map, wmtver, NULL);
      // Request completed
      return MS_SUCCESS;
  }
  else if (request && strcasecmp(request, "GetMap") == 0 &&
           format && strcasecmp(format,  "image/txt") == 0)
  {
      // Until someone adds full support for ASCII graphics this should do. ;)
      printf("Content-type: text/plain\n\n");
      printf(".\n               ,,ggddY\"\"\"Ybbgg,,\n          ,agd888b,_ "
             "\"Y8, ___'\"\"Ybga,\n       ,gdP\"\"88888888baa,.\"\"8b    \""
             "888g,\n     ,dP\"     ]888888888P'  \"Y     '888Yb,\n   ,dP\""
             "      ,88888888P\"  db,       \"8P\"\"Yb,\n  ,8\"       ,8888"
             "88888b, d8888a           \"8,\n ,8'        d88888888888,88P\""
             "' a,          '8,\n,8'         88888888888888PP\"  \"\"      "
             "     '8,\nd'          I88888888888P\"                   'b\n8"
             "           '8\"88P\"\"Y8P'                      8\n8         "
             "   Y 8[  _ \"                        8\n8              \"Y8d8"
             "b  \"Y a                   8\n8                 '\"\"8d,   __"
             "                 8\nY,                    '\"8bd888b,        "
             "     ,P\n'8,                     ,d8888888baaa       ,8'\n '8"
             ",                    888888888888'      ,8'\n  '8a           "
             "        \"8888888888I      a8'\n   'Yba                  'Y88"
             "88888P'    adP'\n     \"Yba                 '888888P'   adY\""
             "\n       '\"Yba,             d8888P\" ,adP\"' \n          '\""
             "Y8baa,      ,d888P,ad8P\"' \n               ''\"\"YYba8888P\""
             "\"''\n");
      return MS_SUCCESS;
  }

  /* If SERVICE, VERSION and REQUEST not included than this isn't a WMS req*/
  if (service == NULL && wmtver==NULL && request==NULL)
      return MS_DONE;  /* Not a WMS request */

  // VERSION *and* REQUEST required by both getMap and getFeatureInfo
  if (wmtver==NULL)
  {
      msSetError(MS_WMSERR,
                 "Incomplete WMS request: VERSION parameter missing",
                 "msWMSDispatch()");
      return msWMSException(map, wmtver, NULL);
  }

  if (request==NULL)
  {
      msSetError(MS_WMSERR,
                 "Incomplete WMS request: REQUEST parameter missing",
                 "msWMSDispatch()");
      return msWMSException(map, wmtver, NULL);
  }

  if ((status = msOWSMakeAllLayersUnique(map)) != MS_SUCCESS)
      return msWMSException(map, wmtver, NULL);

  if (strcasecmp(request, "GetLegendGraphic") == 0)
    return msWMSGetLegendGraphic(map, wmtver, req->ParamNames, req->ParamValues, req->NumParams);

  if (strcasecmp(request, "GetStyles") == 0)
    return msWMSGetStyles(map, wmtver, req->ParamNames, req->ParamValues, req->NumParams);

  // getMap parameters are used by both getMap and getFeatureInfo
  status = msWMSLoadGetMapParams(map, wmtver, req->ParamNames, req->ParamValues, req->NumParams);
  if (status != MS_SUCCESS) return status;

  if (strcasecmp(request, "map") == 0 || strcasecmp(request, "GetMap") == 0)
    return msWMSGetMap(map, wmtver, req->ParamNames, req->ParamValues, req->NumParams);
  else if (strcasecmp(request, "feature_info") == 0 || strcasecmp(request, "GetFeatureInfo") == 0)
    return msWMSFeatureInfo(map, wmtver, req->ParamNames, req->ParamValues, req->NumParams);
  else if (strcasecmp(request, "DescribeLayer") == 0)
  {
      printf("Content-type: text/xml\n\n");
      return msWMSDescribeLayer(map, wmtver, req->ParamNames, req->ParamValues, req->NumParams);
  }

  // Hummmm... incomplete or unsupported WMS request
  msSetError(MS_WMSERR, "Incomplete or unsupported WMS request", "msWMSDispatch()");
  return msWMSException(map, wmtver, NULL);
#else
  msSetError(MS_WMSERR, "WMS server support is not available.", "msWMSDispatch()");
  return(MS_FAILURE);
#endif
}

